from asyncio import AbstractEventLoop
from http import HTTPStatus
from typing import Awaitable, Callable, Optional, Tuple

from sanic import Sanic
from sanic.exceptions import MethodNotSupported, NotFound
from sanic.request import Request
from sanic.response import HTTPResponse

from aiodogstatsd import Client, typedefs
from aiodogstatsd.compat import get_event_loop

__all__ = (
    "DEFAULT_CLIENT_APP_KEY",
    "DEAFULT_REQUEST_DURATION_METRIC_NAME",
    "listeners_factory",
    "middlewares_factory",
)


DEFAULT_CLIENT_APP_KEY = "statsd"
DEAFULT_REQUEST_DURATION_METRIC_NAME = "http_request_duration"


ListenerCallable = Callable[[Sanic, AbstractEventLoop], Awaitable]
MiddlewareCallable = Callable[..., Awaitable]


def listeners_factory(
    *,
    client_app_key: str = DEFAULT_CLIENT_APP_KEY,
    host: str = "localhost",
    port: int = 9125,
    namespace: Optional[typedefs.MNamespace] = None,
    constant_tags: Optional[typedefs.MTags] = None,
    read_timeout: float = 0.5,
    close_timeout: Optional[float] = None,
) -> Tuple[ListenerCallable, ListenerCallable]:
    async def listener_setup(app: Sanic, loop: AbstractEventLoop) -> None:
        client = Client(
            host=host,
            port=port,
            namespace=namespace,
            constant_tags=constant_tags,
            read_timeout=read_timeout,
            close_timeout=close_timeout,
        )
        await client.connect()

        setattr(app, client_app_key, client)

    async def listener_close(app: Sanic, loop: AbstractEventLoop) -> None:
        client = getattr(app, client_app_key)
        await client.close()

    return listener_setup, listener_close


def middlewares_factory(
    *,
    client_app_key: str = DEFAULT_CLIENT_APP_KEY,
    request_duration_metric_name: str = DEAFULT_REQUEST_DURATION_METRIC_NAME,
    collect_not_allowed: bool = False,
    collect_not_found: bool = False,
) -> Tuple[MiddlewareCallable, MiddlewareCallable]:
    async def middleware_request(request: Request) -> None:
        request["_statsd_request_started_at"] = get_event_loop().time()

    async def middleware_response(request: Request, response: HTTPResponse) -> None:
        if _proceed_collecting(
            request, response, collect_not_allowed, collect_not_found
        ):
            request_duration = (
                get_event_loop().time() - request["_statsd_request_started_at"]
            ) * 1000
            getattr(request.app, client_app_key).timing(
                request_duration_metric_name,
                value=request_duration,
                tags={
                    "method": request.method,
                    "path": request.path,
                    "status": response.status,
                },
            )

    return middleware_request, middleware_response


def _proceed_collecting(
    request: Request,
    response: HTTPResponse,
    collect_not_allowed: bool,
    collect_not_found: bool,
) -> bool:
    try:
        request.match_info
    except (MethodNotSupported, NotFound):
        request_match_info_error = True
    else:
        request_match_info_error = False

    if request_match_info_error and (
        (response.status == HTTPStatus.METHOD_NOT_ALLOWED and not collect_not_allowed)
        or (response.status == HTTPStatus.NOT_FOUND and not collect_not_found)
    ):
        return False

    return True



import asyncio
from asyncio.transports import DatagramTransport
from random import random
from typing import Optional

from aiodogstatsd import protocol, typedefs
from aiodogstatsd.compat import get_event_loop

__all__ = ("Client",)


class Client:
    __slots__ = (
        "_host",
        "_port",
        "_namespace",
        "_constant_tags",
        "_closing",
        "_protocol",
        "_queue",
        "_listen_future",
        "_listen_future_join",
        "_read_timeout",
        "_close_timeout",
    )

    def __init__(
        self,
        *,
        host: str = "localhost",
        port: int = 9125,
        namespace: Optional[typedefs.MNamespace] = None,
        constant_tags: Optional[typedefs.MTags] = None,
        read_timeout: float = 0.5,
        close_timeout: Optional[float] = None,
    ) -> None:
        """
        Initialize a client object.

        You can pass host and port of the DogStatsD server, namespace to prefix all
        metric names, constant tags to attach to all metrics.

        Also, you can specify default read timeout which will be used to read messages
        from an AsyncIO queue, and you can specify close timeout which will be used
        as wait time for client closing.
        """
        self._host = host
        self._port = port
        self._namespace = namespace
        self._constant_tags = constant_tags or {}

        self._closing = False

        self._protocol = DatagramProtocol()

        self._queue: asyncio.Queue = asyncio.Queue()
        self._listen_future: asyncio.Future
        self._listen_future_join: asyncio.Future = asyncio.Future()

        self._read_timeout = read_timeout
        self._close_timeout = close_timeout

    async def __aenter__(self) -> "Client":
        await self.connect()
        return self

    async def __aexit__(self, *args) -> None:
        await self.close()

    async def connect(self) -> None:
        loop = get_event_loop()
        await loop.create_datagram_endpoint(
            lambda: self._protocol, remote_addr=(self._host, self._port)
        )

        self._listen_future = asyncio.ensure_future(self._listen())

    async def close(self) -> None:
        self._closing = True

        try:
            await asyncio.wait_for(self._close(), timeout=self._close_timeout)
        except asyncio.TimeoutError:
            pass

    async def _close(self) -> None:
        await self._listen_future_join
        self._listen_future.cancel()

        await self._protocol.close()

    def gauge(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        """
        Record the value of a gauge, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.GAUGE, value, tags, sample_rate)

    def increment(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue = 1,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        """
        Increment a counter, optionally setting a value, tags and a sample rate.
        """
        self._report(name, typedefs.MType.COUNTER, value, tags, sample_rate)

    def decrement(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue = 1,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        """
        Decrement a counter, optionally setting a value, tags and a sample rate.
        """
        value = -value if value else value
        self._report(name, typedefs.MType.COUNTER, value, tags, sample_rate)

    def histogram(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        """
        Sample a histogram value, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.HISTOGRAM, value, tags, sample_rate)

    def distribution(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        """
        Send a global distribution value, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.DISTRIBUTION, value, tags, sample_rate)

    def timing(
        self,
        name: typedefs.MName,
        *,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        """
        Record a timing, optionally setting tags and a sample rate.
        """
        self._report(name, typedefs.MType.TIMING, value, tags, sample_rate)

    async def _listen(self) -> None:
        while not self._closing:
            await self._listen_and_send()

        # Try to send remaining enqueued metrics if any
        while not self._queue.empty():
            await self._listen_and_send()
        self._listen_future_join.set_result(True)

    async def _listen_and_send(self) -> None:
        coro = self._queue.get()

        try:
            buf = await asyncio.wait_for(coro, timeout=self._read_timeout)
        except asyncio.TimeoutError:
            pass
        else:
            self._protocol.send(buf)

    def _report(
        self,
        name: typedefs.MName,
        type_: typedefs.MType,
        value: typedefs.MValue,
        tags: Optional[typedefs.MTags] = None,
        sample_rate: typedefs.MSampleRate = 1,
    ) -> None:
        # If client in closing state, then ignore any new incoming metric
        if self._closing:
            return

        if sample_rate != 1 and random() > sample_rate:
            return

        # Resolve full tags list
        all_tags = dict(self._constant_tags, **tags or {})

        # Build and enqueue metric
        self._queue.put_nowait(
            protocol.build(
                name=name,
                namespace=self._namespace,
                value=value,
                type_=type_,
                tags=all_tags,
                sample_rate=sample_rate,
            )
        )


class DatagramProtocol(asyncio.DatagramProtocol):
    __slots__ = ("_transport", "_closed")

    def __init__(self) -> None:
        self._transport: Optional[DatagramTransport] = None
        self._closed: asyncio.Future = asyncio.Future()

    async def close(self) -> None:
        if self._transport is None:
            return

        self._transport.close()
        await self._closed

    def connection_made(self, transport):
        self._transport = transport

    def connection_lost(self, _exc):
        self._transport = None
        self._closed.set_result(True)

    def send(self, data: bytes) -> None:
        if self._transport is None:
            return

        try:
            self._transport.sendto(data)
        except Exception:
            # Errors should fail silently so they don't affect anything else
            pass



from http import HTTPStatus
from typing import AsyncIterator, Callable, Optional

from aiohttp import web
from aiohttp.web_app import _Middleware
from aiohttp.web_routedef import _SimpleHandler
from aiohttp.web_urldispatcher import MatchInfoError

from aiodogstatsd import Client, typedefs
from aiodogstatsd.compat import get_event_loop

__all__ = (
    "DEFAULT_CLIENT_APP_KEY",
    "DEAFULT_REQUEST_DURATION_METRIC_NAME",
    "cleanup_context_factory",
    "middleware_factory",
)


DEFAULT_CLIENT_APP_KEY = "statsd"
DEAFULT_REQUEST_DURATION_METRIC_NAME = "http_request_duration"


def cleanup_context_factory(
    *,
    client_app_key: str = DEFAULT_CLIENT_APP_KEY,
    host: str = "localhost",
    port: int = 9125,
    namespace: Optional[typedefs.MNamespace] = None,
    constant_tags: Optional[typedefs.MTags] = None,
    read_timeout: float = 0.5,
    close_timeout: Optional[float] = None,
) -> Callable[[web.Application], AsyncIterator[None]]:
    async def cleanup_context(app: web.Application) -> AsyncIterator[None]:
        app[client_app_key] = Client(
            host=host,
            port=port,
            namespace=namespace,
            constant_tags=constant_tags,
            read_timeout=read_timeout,
            close_timeout=close_timeout,
        )
        await app[client_app_key].connect()
        yield
        await app[client_app_key].close()

    return cleanup_context


def middleware_factory(
    *,
    client_app_key: str = DEFAULT_CLIENT_APP_KEY,
    request_duration_metric_name: str = DEAFULT_REQUEST_DURATION_METRIC_NAME,
    collect_not_allowed: bool = False,
    collect_not_found: bool = False,
) -> _Middleware:
    @web.middleware
    async def middleware(
        request: web.Request, handler: _SimpleHandler
    ) -> web.StreamResponse:
        loop = get_event_loop()
        request_started_at = loop.time()

        try:
            response = await handler(request)
            response_status = response.status
        except web.HTTPException as e:
            response_status = e.status
            raise e
        except Exception as e:
            response_status = HTTPStatus.INTERNAL_SERVER_ERROR.value
            raise e
        finally:
            if _proceed_collecting(  # pragma: no branch
                request, response_status, collect_not_allowed, collect_not_found
            ):
                request_duration = (loop.time() - request_started_at) * 1000
                request.app[client_app_key].timing(
                    request_duration_metric_name,
                    value=request_duration,
                    tags={
                        "method": request.method,
                        "path": request.path,
                        "status": response_status,
                    },
                )

        return response

    return middleware


def _proceed_collecting(
    request: web.Request,
    response_status: int,
    collect_not_allowed: bool,
    collect_not_found: bool,
) -> bool:
    if isinstance(request.match_info, MatchInfoError) and (
        (response_status == HTTPStatus.METHOD_NOT_ALLOWED and not collect_not_allowed)
        or (response_status == HTTPStatus.NOT_FOUND and not collect_not_found)
    ):
        return False

    return True



from typing import Optional

from aiodogstatsd import typedefs

__all__ = ("build", "build_tags")


def build(
    *,
    name: typedefs.MName,
    namespace: Optional[typedefs.MNamespace],
    value: typedefs.MValue,
    type_: typedefs.MType,
    tags: typedefs.MTags,
    sample_rate: typedefs.MSampleRate,
) -> bytes:
    p_name = f"{namespace}.{name}" if namespace is not None else name
    p_sample_rate = f"|@{sample_rate}" if sample_rate != 1 else ""

    p_tags = build_tags(tags)
    p_tags = f"|#{p_tags}" if p_tags else ""

    return f"{p_name}:{value}|{type_.value}{p_sample_rate}{p_tags}".encode("utf-8")


def build_tags(tags: typedefs.MTags) -> str:
    if not tags:
        return ""

    return ",".join(f"{k}:{v}" for k, v in tags.items())
